! WHIZARD 2.2.3 Aug 15 2014
! 
! Copyright (C) 1999-2014 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions from
!     Fabian Bach <fabian.bach@desy.de>
!     Christian Speckner <cnspeckn@googlemail.com> 
!     Christian Weiss <christian.weiss@desy.de>
!     and Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module prc_gosam
  
  use iso_c_binding !NODEP!
  use kinds, only: default, double !NODEP!
  use file_utils !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!  
  use constants !NODEP!
  use diagnostics !NODEP!
  use limits, only: TAB !NODEP!  
  use os_interface
  use interactions
  use lorentz !NODEP!
  use system_dependencies !NODEP!
  use pdg_arrays
  use sm_qcd
  use flavors
  use models
  use md5
  
  use process_constants
  use prclib_interfaces
  use prc_core_def
  use prc_core

  use blha_config

  implicit none
  private

  character(10), dimension(5), parameter ::  &
             lib_suffix = [character(10) :: &
             '.a', '.la', '.so', '.so.0', '.so.0.0.0']

  public :: gosam_writer_t
  public :: gosam_def_t
  public :: prc_gosam_t

  type, extends (prc_writer_f_module_t) :: gosam_writer_t
     ! integer, dimension(:), allocatable :: pdg_in
     ! integer, dimension(:), allocatable :: pdg_out
    type(string_t) :: gosam_dir
    type(string_t) :: golem_dir
    type(string_t) :: samurai_dir
    type(string_t) :: ninja_dir
    type(string_t) :: form_dir
    type(string_t) :: qgraf_dir
    type(blha_configuration_t) :: blha_cfg
  contains
    procedure :: write_wrapper => gosam_write_wrapper
    procedure :: write_interface => gosam_write_interface
    procedure :: write_source_code => gosam_write_source_code
    procedure :: write_makefile_code => gosam_write_makefile_code
    procedure, nopass :: type_name => gosam_writer_type_name
    procedure :: write => gosam_writer_write
    procedure :: init => gosam_writer_init
    procedure :: generate_configuration_file => &
              gosam_writer_generate_configuration_file
    procedure :: generate_olp_file => gosam_writer_generate_olp_file
    procedure :: get_process_string => gosam_writer_get_process_string
    procedure :: get_n_proc => gosam_writer_get_n_proc
  end type gosam_writer_t

  type, extends (prc_core_def_t) :: gosam_def_t
    type(string_t) :: basename
    logical :: execute_olp = .true.
  contains
    procedure :: init => gosam_def_init
    procedure, nopass :: type_string => gosam_def_type_string
    procedure :: write => gosam_def_write
    procedure :: read => gosam_def_read
    procedure :: allocate_driver => gosam_def_allocate_driver
    procedure, nopass :: get_features => gosam_def_get_features
    procedure :: connect => gosam_def_connect
  end type gosam_def_t

  type, extends (prc_core_driver_t) :: gosam_driver_t 
     ! type(string_t) :: gosam_srcdir
     ! type(string_t) :: gosam_bindir
     ! type(string_t) :: golem_srcdir
     ! type(string_t) :: samurai_srcdir
     ! type(string_t) :: ninja_srcdir
     ! type(string_t) :: form_bindir
    type(string_t) :: gosam_dir
    type(string_t) :: olp_file
    type(string_t) :: olc_file
    type(string_t) :: olp_dir
    procedure(olp_start),nopass,  pointer :: &
              gosam_olp_start => null ()
    procedure(olp_eval), nopass, pointer :: &
              gosam_olp_eval => null()
    procedure(olp_info), nopass, pointer :: &
              gosam_olp_info => null ()
    procedure(olp_set_parameter), nopass, pointer :: &
              gosam_olp_set_parameter => null ()
    procedure(olp_eval2), nopass, pointer :: &
              gosam_olp_eval2 => null ()
    procedure(olp_option), nopass, pointer :: &
              gosam_olp_option => null ()
    procedure(olp_polvec), nopass, pointer :: &
              gosam_olp_polvec => null ()
    procedure(olp_finalize), nopass, pointer :: &
              gosam_olp_finalize => null ()
    procedure(olp_print_parameter), nopass, pointer :: &
              gosam_olp_print_parameter => null ()
  contains
    procedure, nopass :: type_name => gosam_driver_type_name
    procedure :: init => gosam_driver_init
    ! procedure :: setup_gosam => gosam_driver_setup_gosam
    !  procedure :: install_golem => gosam_driver_install_golem
    !  procedure :: install_samurai => gosam_driver_install_samurai
    !  procedure :: install_ninja => gosam_driver_install_ninja
    !  procedure :: unpack_form => gosam_driver_unpack_form
    procedure :: execute_olp_file => gosam_driver_execute_olp_file
    procedure :: load => gosam_driver_load
  end type gosam_driver_t

  type, extends (prc_core_t) :: prc_gosam_t
    type(qcd_t) :: qcd
    integer :: n_proc
    real(default) :: maximum_accuracy = 10000.0
  contains
    procedure :: write => prc_gosam_write
    procedure :: needs_mcset => prc_gosam_needs_mcset
    procedure :: get_n_terms => prc_gosam_get_n_terms
    procedure :: is_allowed => prc_gosam_is_allowed
    procedure :: compute_hard_kinematics => prc_gosam_compute_hard_kinematics
    procedure :: compute_eff_kinematics => prc_gosam_compute_eff_kinematics
    procedure :: compute_amplitude => prc_gosam_compute_amplitude
    procedure :: recover_kinematics => prc_gosam_recover_kinematics
     procedure :: init_writer => prc_gosam_init_writer
    procedure :: init_driver => prc_gosam_init_driver
    procedure :: set_parameters => prc_gosam_set_parameters
    procedure :: compute_sqme_fin => prc_gosam_compute_sqme_fin
    procedure :: fill_constants => prc_gosam_fill_constants
    procedure :: allocate_workspace => prc_gosam_allocate_workspace
    procedure :: get_alpha_s => prc_gosam_get_alpha_s
    procedure :: set_n_proc => prc_gosam_set_n_proc
    procedure :: get_n_proc => prc_gosam_get_n_proc
  end type prc_gosam_t

  type, extends (workspace_t) :: gosam_state_t
    logical :: new_kinematics = .true.
    real(default) :: alpha_qcd = -1
  contains
    procedure :: write => gosam_state_write
    procedure :: reset_new_kinematics => gosam_state_reset_new_kinematics
  end type gosam_state_t


  interface 
    subroutine olp_start (contract_file_name, ierr) bind (C,name="OLP_Start")
      import
      character(kind=c_char, len=1), intent(in) :: contract_file_name
      integer(kind=c_int), intent(out) :: ierr
    end subroutine olp_start
  end interface

  interface
    subroutine olp_eval (label, momenta, mu, parameters, res) &
         bind (C,name="OLP_EvalSubProcess")
      import
      integer(kind=c_int), value, intent(in) :: label
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(50), intent(in) :: momenta
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(60), intent(out) :: res
    end subroutine olp_eval
  end interface

  interface
    subroutine olp_info (olp_file, olp_version, message) &
         bind(C,name="OLP_Info")
      import
      character(kind=c_char), intent(inout), dimension(15) :: olp_file
      character(kind=c_char), intent(inout), dimension(15) :: olp_version
      character(kind=c_char), intent(inout), dimension(255) :: message
    end subroutine olp_info
  end interface

  interface
    subroutine olp_set_parameter &
         (variable_name, real_part, complex_part, success) &
            bind(C,name="OLP_SetParameter")
      import
      character(kind=c_char,len=1), intent(in) :: variable_name
      real(kind=c_double), intent(in) :: real_part, complex_part
      integer(kind=c_int), intent(out) :: success
    end subroutine olp_set_parameter
  end interface

  interface
    subroutine olp_eval2 (label, momenta, mu, res, acc) &
         bind(C,name="OLP_EvalSubProcess2")
      import
      integer(kind=c_int), intent(in) :: label
      real(kind=c_double), intent(in) :: mu
      real(kind=c_double), dimension(50), intent(in) :: momenta
      real(kind=c_double), dimension(60), intent(out) :: res
      real(kind=c_double), intent(out) :: acc
    end subroutine olp_eval2
  end interface

  interface
    subroutine olp_option (line, stat) bind(C,name="OLP_Option")
      import
      character(kind=c_char, len=1), intent(in) :: line
      integer(kind=c_int), intent(out) :: stat
    end subroutine
  end interface

  interface
    subroutine olp_polvec (p, q, eps) bind(C,name="OLP_Polvec")
      import
      real(kind=c_double), dimension(0:3), intent(in) :: p, q
      real(kind=c_double), dimension(0:7), intent(out) :: eps
    end subroutine
  end interface

  interface
    subroutine olp_finalize () bind(C,name="OLP_Finalize")
      import
    end subroutine olp_finalize
  end interface

  interface
    subroutine olp_print_parameter (filename) bind(C,name="OLP_PrintParameter")
      import
      character(kind=c_char, len=1), intent(in) :: filename
    end subroutine olp_print_parameter
  end interface


contains

  subroutine gosam_def_init (object, basename)
    class(gosam_def_t), intent(inout) :: object
    type(string_t), intent(in) :: basename
    object%basename = basename
    allocate (gosam_writer_t :: object%writer)
  end subroutine gosam_def_init

  function gosam_def_type_string () result (string)
    type(string_t) :: string
    string = "gosam"
  end function gosam_def_type_string

  subroutine gosam_def_write (object, unit)
    class(gosam_def_t), intent(in) :: object
    integer, intent(in) :: unit
    select type (writer => object%writer)
    type is (gosam_writer_t)
      call writer%write (unit)
    end select
  end subroutine gosam_def_write

  subroutine gosam_def_read (object, unit)
    class(gosam_def_t), intent(out) :: object
    integer, intent(in) :: unit
    call msg_bug ("GoSam process definition: input not supported yet")
  end subroutine gosam_def_read


  subroutine gosam_def_allocate_driver (object, driver, basename)
    class(gosam_def_t), intent(in) :: object
    class(prc_core_driver_t), intent(out), allocatable :: driver
    type(string_t), intent(in) :: basename
    if (.not. allocated (driver)) allocate (gosam_driver_t :: driver)
  end subroutine gosam_def_allocate_driver

  subroutine gosam_def_get_features (features)
    type(string_t), dimension(:), allocatable, intent(out) :: features
    allocate (features (6))
    features = [ &
         var_str ("init"), &
         var_str ("update_alpha_s"), &
         var_str ("reset_helicity_selection"), &
         var_str ("is_allowed"), &
         var_str ("new_event"), &
         var_str ("get_amplitude")]
  end subroutine gosam_def_get_features 

  subroutine gosam_def_connect (def, lib_driver, i, proc_driver)   
    class(gosam_def_t), intent(in) :: def
    class(prclib_driver_t), intent(in) :: lib_driver
    integer, intent(in) :: i
    class(prc_core_driver_t), intent(inout) :: proc_driver
    call msg_bug ("No implementation for gosam_def_connect")
  end subroutine gosam_def_connect

  subroutine gosam_write_wrapper (writer, unit, id, feature)
    class(gosam_writer_t), intent(in) :: writer
    integer, intent(in) :: unit
    type(string_t), intent(in) :: id, feature
    call msg_bug ("Gosam write wrapper: Do not know what to do!")
  end subroutine gosam_write_wrapper

  subroutine gosam_write_interface (writer, unit, id, feature)
    class(gosam_writer_t), intent(in) :: writer
    integer, intent(in) :: unit
    type(string_t), intent(in) :: id, feature
  end subroutine gosam_write_interface 

  subroutine gosam_write_source_code (writer, id)
    class(gosam_writer_t), intent(in) :: writer
    type(string_t), intent(in) :: id
    call msg_bug ("Gosam does not have to be " // &
         "supported with Whizard-generated code")
  end subroutine gosam_write_source_code

  subroutine gosam_write_makefile_code (writer, unit, id, os_data, testflag)
    class(gosam_writer_t), intent(in) :: writer
    integer, intent(in) :: unit
    type(string_t), intent(in) :: id
    type(os_data_t), intent(in) :: os_data
    logical, intent(in), optional :: testflag
    call msg_bug ("Gosam does not have to be " // &
         "supported with a Whizard-generated Makefile")
  end subroutine gosam_write_makefile_code

  function gosam_writer_type_name () result (string)
    type(string_t) :: string
    string = "gosam"
  end function gosam_writer_type_name

  subroutine gosam_writer_write (writer, unit)
    class(gosam_writer_t), intent(in) :: writer
    integer, intent(in) :: unit    
    write (unit, "(1x,A)")  char (writer%get_process_string ())
  end subroutine gosam_writer_write

  subroutine gosam_writer_init (writer, basename, flv_states, os_data, &
                                model, ex_olp)
    class(gosam_writer_t), intent(inout) :: writer
    type(string_t), intent(in) :: basename
    integer, intent(in), dimension(:,:) :: flv_states
    type(os_data_t), intent(in) :: os_data
    type(model_t), intent(in), pointer :: model
    logical, intent(out) :: ex_olp
    ex_olp = .false.

    writer%gosam_dir = GOSAM_DIR
    writer%golem_dir = GOLEM_DIR 
    writer%samurai_dir = SAMURAI_DIR
    writer%ninja_dir = NINJA_DIR
    writer%form_dir = FORM_DIR
    writer%qgraf_dir = QGRAF_DIR

    call blha_configuration_init (writer%blha_cfg, basename, model, &
                                  BLHA_MODE_GOSAM)
    call blha_configuration_append_processes (writer%blha_cfg, &
                                              flv_states)
    call blha_configuration_set (writer%blha_cfg, 2, &
                                    amplitude_type = BLHA_AMP_LOOP, &
                                    correction_type = BLHA_CT_QCD, &
                                    irreg = BLHA_IRREG_CDR, &
                                    alphas_power = 0, &
                                    alpha_power = 2, &
                                    ew_scheme = BLHA_EW_GF, &
                                    debug = .true.)
    call check_file_change (var_str ('golem.in'), 'config')
    call check_file_change (basename // '.olp', 'olp')
  contains
    subroutine check_file_change (filename, file_type)
      type(string_t), intent(in) :: filename
      character(*), intent(in) :: file_type
      logical :: exist
      integer :: unit
      character(len=32) :: md5sum1, md5sum2
      inquire (file = char (filename), exist = exist)
      if (exist) then
        unit = free_unit ()
        open (unit, file = char (filename), status = 'old', action = 'read')
        md5sum1 = md5sum (unit)
        close (unit)
        open (unit, file = char (filename // '_tmp'), status = 'replace',&
              action = 'readwrite')
        call generate_file (unit, file_type)
        rewind (unit)
        md5sum2 = md5sum (unit)
        close (unit)
        if (md5sum1 == md5sum2) then
          call os_system_call ('rm ' // filename // '_tmp')
          ex_olp = ex_olp .or. .false.
        else
          call os_system_call ('mv ' // filename // '_tmp' &
                  // ' ' // filename)
          ex_olp = ex_olp .or. .true.
        end if
      else
        unit = free_unit ()
        open (unit, file = char (filename), status = 'new', action = 'write')
        call generate_file (unit, file_type)
        close(unit)
        ex_olp = ex_olp .or..true.
      end if
    end subroutine check_file_change

    subroutine generate_file (unit, file_type)
      integer, intent(in) :: unit
      character(*), intent(in) :: file_type
      select case (file_type)
      case ('config')
        call writer%generate_configuration_file (unit, os_data)
      case ('olp')
        call writer%generate_olp_file (unit, os_data)
      case default
        call msg_fatal &
          ('Gosam writer: can only create configuration- and olp-files')
      end select
    end subroutine generate_file
  end subroutine gosam_writer_init  

  function gosam_driver_type_name () result (string)
    type(string_t) :: string
    string = "gosam"
  end function gosam_driver_type_name

  subroutine gosam_driver_init (object, os_data, olp_file, &
                                olc_file, olp_dir)
    class(gosam_driver_t), intent(inout) :: object
    type(os_data_t), intent(in) :: os_data
    type(string_t), intent(in) :: olp_file, olc_file, olp_dir
    ! object%gosam_srcdir = GOSAM_DIR
    ! object%gosam_bindir = os_data%prefix // '/bin'
    ! object%golem_srcdir = GOLEM_DIR
    ! object%samurai_srcdir = SAMURAI_DIR
    ! object%ninja_srcdir = NINJA_DIR
    ! object%form_bindir = FORM_DIR
    object%gosam_dir = GOSAM_DIR
    object%olp_file = olp_file
    object%olc_file = olc_file
    object%olp_dir = olp_dir
  end subroutine gosam_driver_init

  subroutine gosam_writer_generate_configuration_file &
          (object, unit, os_data)
      class(gosam_writer_t), intent(in) :: object
      integer, intent(in) :: unit
      type(os_data_t), intent(in) :: os_data
      type(string_t) :: fc_bin
      type(string_t) :: form_bin, qgraf_bin, haggies_bin
      type(string_t) :: fcflags_golem, ldflags_golem
      type(string_t) :: fcflags_samurai, ldflags_samurai
      type(string_t) :: fcflags_ninja, ldflags_ninja
      type(string_t) :: ldflags_avh_olo, ldflags_qcdloop
      fc_bin = DEFAULT_FC
      ! form_bindir = '/usr/local/bin/tform'
      ! form_bindir = os_data%whizard_omega_binpath // '/tform'
      ! qgraf_bindir = os_data%whizard_omega_binpath // '/qgraf'
      form_bin = object%form_dir // '/bin/tform'
      qgraf_bin = object%qgraf_dir // '/bin/qgraf'
      if (object%gosam_dir /= "") then
        haggies_bin = '/usr/bin/java -jar ' // object%gosam_dir // &
                       '/share/golem/haggies/haggies.jar'
      else
        call msg_fatal ("generate_configuration_file: At least " // &
             "the GoSam Directory has to be specified!")
      end if
      if (object%golem_dir /= "") then
        fcflags_golem = "-I" // object%golem_dir // "/include/golem95"
        ldflags_golem = "-L" // object%golem_dir // "/lib -lgolem"
      end if
      if (object%samurai_dir /= "") then
        fcflags_samurai = "-I" // object%samurai_dir // "/include/samurai"
        ldflags_samurai = "-L" // object%samurai_dir // "/lib -lsamurai"
        ldflags_avh_olo = "-L" // object%samurai_dir // "/lib -lavh_olo"
        ldflags_qcdloop = "-L" // object%samurai_dir // "/lib -lqcdloop"
      end if
      ! fcflags_ninja = "-I" // os_data%prefix // "/include/ninja " &
      !                  // "-I" // os_data%prefix // "/include"
      ! ldflags_ninja = "-L" // os_data%prefix // "/lib -lninja"
      if (object%ninja_dir /= "") then
        fcflags_ninja = "-I" // object%ninja_dir // "/include/ninja " &
                        // "-I" // object%ninja_dir // "/include"
        ldflags_ninja = "-L" // object%ninja_dir // "/lib -lninja"
      end if
      write (unit, "(A)") "+avh_olo.ldflags=" &
            // char (ldflags_avh_olo) 
      ! write (unit, "(A)") "reduction_programs=golem95, samurai"
      write (unit, "(A)") "reduction_programs=golem95, samurai, ninja"
      ! write (unit, "(A)") "extensions=noformopt, autotools"
      write (unit, "(A)") "extensions=autotools"
      write (unit, "(A)") "+qcdloop.ldflags=" &
            // char (ldflags_qcdloop)
      write (unit, "(A)") "+zzz.extensions=qcdloop, avh_olo"
      write (unit, "(A)") "fc.bin=" // char (os_data%fc)
      write (unit, "(A)") "form.bin=" // char (form_bin)
      write (unit, "(A)") "qgraf.bin=" // char (qgraf_bin)
      write (unit, "(A)") "golem95.fcflags=" // char (fcflags_golem)
      write (unit, "(A)") "golem95.ldflags=" // char (ldflags_golem)
      write (unit, "(A)") "haggies.bin=" // char (haggies_bin)
      write (unit, "(A)") "samurai.fcflags=" // char (fcflags_samurai)
      write (unit, "(A)") "samurai.ldflags=" // char (ldflags_samurai)
      write (unit, "(A)") "ninja.fcflags=" // char (fcflags_ninja)
      write (unit, "(A)") "ninja.ldflags=" // char (ldflags_ninja)
      write (unit, "(A)") "zero=mU,mD,mC,mS,mB"
      write (unit, "(A)") "PSP_check=False"
      write (unit, "(A)") "filter.lo=lambda d: d.iprop(H) == 0 and d.iprop(chi) == 0"
      write (unit, "(A)") "filter.nlo=lambda d: d.iprop(H) == 0 and d.iprop(chi) == 0"
  end subroutine gosam_writer_generate_configuration_file

  subroutine gosam_writer_generate_olp_file (object, unit, os_data)
    class(gosam_writer_t), intent(in) :: object
    integer, intent(in) :: unit
    type(os_data_t), intent(in) :: os_data
    call blha_configuration_write (object%blha_cfg, unit)
    ! write (unit, "(A)") "InterfaceVersion BLHA2"
    ! write (unit, "(A,1X,A)") "MatrixElementSquareType", "CHsummed"
    ! write (unit, "(A,1X,A)") "MatrixElementSquareType", "CHaveraged"
    ! write (unit, "(A,1X,A)") "IRregularisation", "tHV"
    ! write (unit, "(A,1X,A)") "#OperationMode", "CouplingsStrippedOff"
    ! write (unit, "(A,1X,A)") "AlphasPower", "0"
    ! write (unit, "(A,1X,A)") "AlphaPower", "2"
    ! write (unit, "(A,1X,A)") "CorrectionType", "QCD"
    ! write (unit, "(A)") "EWScheme alphaGF"
    ! write (unit, "(A)")  char (object%get_process_string ())
  end subroutine gosam_writer_generate_olp_file

  function gosam_writer_get_process_string (writer) result (s_proc)
    class(gosam_writer_t), intent(in) :: writer
    type(string_t) :: s_proc
    character (3) :: c 
    integer :: i
    ! write (c, '(I3)') writer%pdg_in (1)
    ! s_proc = var_str (c) // " "
    ! write (c, '(I3)') writer%pdg_in (2)
    ! s_proc = s_proc // var_str (c) // " -> "
    ! do i = 1, size (writer%pdg_out)
    !   write (c, '(I3)') writer%pdg_out (i) 
    !   s_proc = s_proc // var_str (c) // " "
    ! end do
  end function gosam_writer_get_process_string

  function gosam_writer_get_n_proc (writer) result (n_proc)
    class(gosam_writer_t), intent(in) :: writer
    integer :: n_proc
    n_proc = blha_configuration_get_n_proc (writer%blha_cfg)
  end function gosam_writer_get_n_proc

  !  subroutine gosam_driver_setup_gosam (object, os_data)
  !        class(gosam_driver_t), intent(in) :: object
  !        type(os_data_t), intent(in) :: os_data
  !        type(string_t) :: command
  !        call msg_message ('Checking for Gosam installation: ')
  !        if (.not. os_file_exist (object%gosam_dir // '/bin/gosam.py')) then
  !          call msg_message ('Gosam.py not present -> Perform installation')
  !          command = "python " // object%gosam_dir // "/src/setup.py install --prefix=" &
  !                    // os_data%prefix
  !          command = "cd " // object%gosam_dir // "/src" // &
  !               "; python setup.py install --prefix=" // os_data%prefix
  !          call os_system_call (command, verbose = .true.)
  !        else
  !          call msg_message ('Gosam already installed')
  !        end if
  !  end subroutine gosam_driver_setup_gosam
  !
  !  subroutine gosam_driver_install_golem (object, os_data)
  !        class(gosam_driver_t), intent(in) :: object
  !        type(os_data_t), intent(in) :: os_data
  !        type(string_t) :: command
  !        call msg_message ('Checking for Golem installation: ')
  !        if (.not. check_golem_installation (os_data)) then
  !          call msg_message ('Golem libraries are not present -> Perform Installation')
  !          command = object%golem_dir // '/src/configure --prefix=' // os_data%prefix
  !          call os_system_call (command, verbose = .true.)
  !          call os_system_call (var_str ("make") , verbose = .true.)
  !          call os_system_call (var_str ("make install"), verbose = .true.)
  !        else
  !          call msg_message ('Golem libraries are present -> Skip Installation')
  !        end if
  !  end subroutine gosam_driver_install_golem
  !
  !  subroutine gosam_driver_install_samurai (object, os_data)
  !        class(gosam_driver_t), intent(in) :: object
  !        type(os_data_t), intent(in) :: os_data
  !        type(string_t) :: command
  !        call msg_message ('Checking for Samurai installation: ')
  !        if (.not. check_samurai_installation (os_data)) then
  !          call msg_message ('Samurai libraries are not present -> Perform Installation')
  !          command = object%samurai_dir // "/src/configure --prefix=" // os_data%prefix
  !          call os_system_call (command, verbose = .true.)
  !          call os_system_call (var_str ("make"), verbose = .true.)
  !          call os_system_call (var_str ("make install"), verbose = .true.)
  !        else
  !          call msg_message ('Samurai libraries are present -> Skip installation')
  !        end if
  !  end subroutine gosam_driver_install_samurai
  !
  !  subroutine gosam_driver_install_ninja (object, os_data)
  !    class(gosam_driver_t), intent(in) :: object
  !    type(os_data_t), intent(in) :: os_data
  !    type(string_t) :: command
  !    call msg_message ('Checking for Ninja installation: ')
  !    if (.not. check_ninja_installation (os_data)) then
  !      call msg_message ('Ninja libraries are not present -> Perform Installation')
  !      command = object%ninja_dir // '/src/configure --prefix=' // os_data%prefix &
  !            // " --with-avholo='-L" // os_data%prefix // "/lib -lavh_olo' " &
  !            // "FCINCLUDE=-I" // os_data%prefix // "/include/golem95"
  !      call os_system_call (command, verbose = .true.)
  !      call os_system_call (var_str ("make"), verbose = .true.)
  !      call os_system_call (var_str ("make install"), verbose = .true.)
  !    else
  !      call msg_message ('Ninja libraries are present -> Skip installation')
  !    end if
  !  end subroutine gosam_driver_install_ninja
  !
  !  subroutine gosam_driver_unpack_form (object, os_data)
  !    class(gosam_driver_t), intent(in) :: object
  !    type(os_data_t), intent(in) :: os_data
  !    type(string_t) :: form_basename, command  
  !    logical :: form_exist, tform_exist
  !    inquire (file = char (os_data%whizard_omega_binpath // "/form"), &
  !             exist = form_exist)
  !    inquire (file = char (os_data%whizard_omega_binpath // "/tform"), &
  !             exist = tform_exist)
  !    if (.not. (form_exist .and. tform_exist)) then
  !      call msg_message ("Form binaries are not present -> Unpack from source")
  !      form_basename = "form-4.1-x86_64-linux"
  !      command = "tar -xf " // object%form_bindir // "/" &
  !              // form_basename // ".tar.gz -C " &
  !              // os_data%whizard_omega_binpath
  !      call os_system_call (command, verbose = .true.)
  !      command = "cd " // os_data%whizard_omega_binpath // "; " &
  !              // "mv " // form_basename // "/form . ;" &
  !              // "mv " // form_basename // "/tform . ;" &
  !              // "rmdir " // form_basename
  !      call os_system_call (command, verbose = .true.)
  !    else
  !      call msg_message ("Form binaries are present")
  !    end if
  !  end subroutine gosam_driver_unpack_form
  !
  subroutine gosam_driver_execute_olp_file (object, os_data)
    class(gosam_driver_t), intent(in) :: object
    type(os_data_t), intent(in) :: os_data
    type(string_t) :: command
    integer :: unit
    ! command = os_data%whizard_omega_binpath // & 
    !          '/gosam.py --olp ' // object%olp_file &
    !          // ' --destination=' // object%olp_dir &
    !          // ' -f -z' 
    command = object%gosam_dir // "/bin" // &
              '/gosam.py --olp ' // object%olp_file // &
              ' --destination=' // object%olp_dir // &
              ' -f -z'
    call os_system_call (command, verbose = .true.)
    call msg_message ("Configure olp_modules")
    command = "cd " // object%olp_dir // "; ./autogen.sh --prefix=$(pwd)"
    call os_system_call (command, verbose = .true.)
    call msg_message ("Installing olp_modules")
    command = "cd " // object%olp_dir // ";  make install"
    call os_system_call (command, verbose = .true.)
  end subroutine gosam_driver_execute_olp_file

  subroutine gosam_driver_load (object, os_data, success)
    class(gosam_driver_t), intent(inout) :: object
    type(os_data_t), intent(in) :: os_data
    logical, intent(out) :: success
    type(dlaccess_t) :: dlaccess
    type(string_t) ::  path
    type(c_funptr) :: c_fptr
    path = object%olp_dir // '/.libs'
    call dlaccess_init &
          (dlaccess, var_str ("."), path // '/libgolem_olp.so', os_data)
    if (os_file_exist (path // '/libgolem_olp.so')) then
      c_fptr = dlaccess_get_c_funptr (dlaccess, var_str ("OLP_Start"))
      call c_f_procpointer (c_fptr, object%gosam_olp_start)

      c_fptr = dlaccess_get_c_funptr (dlaccess, var_str ("OLP_EvalSubProcess"))
      call c_f_procpointer (c_fptr, object%gosam_olp_eval)

      c_fptr = dlaccess_get_c_funptr (dlaccess, var_str ("OLP_Info"))
      call c_f_procpointer (c_fptr, object%gosam_olp_info)

      c_fptr = dlaccess_get_c_funptr (dlaccess, var_str ("OLP_SetParameter"))
      call c_f_procpointer (c_fptr, object%gosam_olp_set_parameter)

      c_fptr = dlaccess_get_c_funptr (dlaccess, var_str ("OLP_EvalSubProcess2"))
      call c_f_procpointer (c_fptr, object%gosam_olp_eval2)

      c_fptr = dlaccess_get_c_funptr (dlaccess, var_str ("OLP_Option"))
      call c_f_procpointer (c_fptr, object%gosam_olp_option)

      c_fptr = dlaccess_get_c_funptr (dlaccess, var_str ("OLP_Polvec"))
      call c_f_procpointer (c_fptr, object%gosam_olp_polvec)

      c_fptr = dlaccess_get_c_funptr (dlaccess, var_str ("OLP_Finalize"))
      call c_f_procpointer (c_fptr, object%gosam_olp_finalize)

      c_fptr = dlaccess_get_c_funptr (dlaccess, var_str ("OLP_PrintParameter"))
      call c_f_procpointer (c_fptr, object%gosam_olp_print_parameter)

      success = .not. dlaccess_has_error (dlaccess)
    else
      success = .false.
    end if
  end subroutine gosam_driver_load

  subroutine prc_gosam_write (object, unit)
    class(prc_gosam_t), intent(in) :: object
    integer, intent(in), optional :: unit
    ! call msg_bug ("prc_gosam_write: Not yet decided what to do!")
    call msg_message ("GOSAM")
  end subroutine prc_gosam_write

  function prc_gosam_needs_mcset (object) result (flag)
    class(prc_gosam_t), intent(in) :: object
    logical :: flag
    !!! Is this really the same for Gosam as for Omega?
    flag = .true.
  end function prc_gosam_needs_mcset

  function prc_gosam_get_n_terms (object) result (n)
    class(prc_gosam_t), intent(in) :: object
    integer :: n
    !!! Again: Identical code for Gosam?
    n = 1
  end function prc_gosam_get_n_terms

  function prc_gosam_is_allowed (object, i_term, f, h, c) result (flag)
    class(prc_gosam_t), intent(in) :: object
    integer, intent(in) :: i_term, f, h, c
    logical :: flag
    ! call msg_bug ("prc_gosam_is_allowed: Do not know how to " // &
    !     "find out which configurations are allowed!")  
    flag = .true.
  end function prc_gosam_is_allowed

  subroutine prc_gosam_compute_hard_kinematics &
       (object, p_seed, i_term, int_hard, tmp)
    class(prc_gosam_t), intent(in) :: object
    type(vector4_t), dimension(:), intent(in) :: p_seed
    integer, intent(in) :: i_term
    type(interaction_t), intent(inout) :: int_hard
    class(workspace_t), intent(inout), allocatable :: tmp 
    call interaction_set_momenta (int_hard, p_seed)
    if (allocated (tmp)) then
      select type (tmp)
      type is (gosam_state_t); tmp%new_kinematics = .true.
      end select
    end if
  end subroutine prc_gosam_compute_hard_kinematics

  subroutine prc_gosam_compute_eff_kinematics &
       (object, i_term, int_hard, int_eff, tmp)
    class(prc_gosam_t), intent(in) :: object
    integer, intent(in) :: i_term
    type(interaction_t), intent(in) :: int_hard
    type(interaction_t), intent(inout) :: int_eff
    class(workspace_t), intent(inout), allocatable :: tmp
  end subroutine prc_gosam_compute_eff_kinematics

  function prc_gosam_compute_amplitude &
       (object, j, p, f, h, c, fac_scale, ren_scale, tmp) result (amp)
    class(prc_gosam_t), intent(in) :: object
    integer, intent(in) :: j
    type(vector4_t), dimension(:), intent(in) :: p
    integer, intent(in) :: f, h, c
    real(default), intent(in) :: fac_scale, ren_scale
    class(workspace_t), intent(inout), allocatable, optional :: tmp
    complex(default) :: amp
    select type (tmp)
    type is (gosam_state_t)
      tmp%alpha_qcd = object%qcd%alpha%get (fac_scale)
    end select
    amp = 0.0
  end function prc_gosam_compute_amplitude

  subroutine prc_gosam_recover_kinematics &
       (object, p_seed, int_hard, int_eff, tmp)
    class(prc_gosam_t), intent(in) :: object
    type(vector4_t), dimension(:), intent(inout) :: p_seed
    type(interaction_t), intent(inout) :: int_hard, int_eff
    class(workspace_t), intent(inout), allocatable :: tmp
    integer :: n_in
    n_in = interaction_get_n_in (int_eff)
    call interaction_set_momenta (int_eff, p_seed(1:n_in), outgoing = .false.)
    p_seed(n_in+1:) = interaction_get_momenta (int_eff, outgoing = .true.)
  end subroutine prc_gosam_recover_kinematics

  subroutine prc_gosam_init_writer (object, prc_const, os_data, model)
    class(prc_gosam_t), intent(in) :: object
    type(process_constants_t), intent(inout) :: prc_const
    type(os_data_t), intent(in) :: os_data
    type(model_t), intent(in), pointer :: model
    integer, dimension(:,:), allocatable :: flv_states
    call prc_const%get_flv_state (flv_states)
    select type (def => object%def)
    type is (gosam_def_t)
      select type (writer => def%writer)
      type is (gosam_writer_t)
        call writer%init (def%basename, flv_states, os_data, model, &
                               def%execute_olp)
      end select
    end select
  end subroutine prc_gosam_init_writer

  subroutine prc_gosam_init_driver (object, os_data)
    class(prc_gosam_t), intent(inout) :: object
    type(os_data_t), intent(in) :: os_data
    logical :: dl_success
    type(string_t) :: olp_file, olc_file, olp_dir
    integer(c_int) :: success
    select type (def => object%def)
    type is (gosam_def_t)
      olp_file = def%basename // '.olp'
      olc_file = def%basename // '.olc'
      olp_dir = def%basename // '_olp_modules'
    class default
      call msg_bug ("prc_gosam_init_driver: core_def should be of gosam-type")
    end select
    call object%def%allocate_driver (object%driver, var_str ("Test_Gosam"))
    associate (driver=>object%driver)
    select type(driver)
    type is (gosam_driver_t)
      call driver%init (os_data, olp_file, olc_file, olp_dir)
      ! call driver%install_golem (os_data)
      ! call driver%install_samurai (os_data)
      ! call driver%install_ninja (os_data)
      ! call driver%unpack_form (os_data)
      ! call driver%setup_gosam (os_data)
      ! call driver%write_installation_script (os_data)
      select type (def => object%def)
      type is (gosam_def_t)
        if (def%execute_olp) &
          call driver%execute_olp_file (os_data)
      end select  
      call driver%load (os_data, dl_success)
      if (.not. dl_success) &
        call msg_fatal ("Error in linking gosam libraries")
      call driver%gosam_olp_start (char (driver%olc_file), success)
    end select
    end associate
  end subroutine prc_gosam_init_driver

  subroutine prc_gosam_set_parameters (object, qcd)
    class(prc_gosam_t), intent(inout) :: object
    type(qcd_t), intent(in) :: qcd
    object%qcd = qcd
  end subroutine prc_gosam_set_parameters

  subroutine prc_gosam_compute_sqme_fin &
       (object, i_proc, p, ren_scale, alpha_s, sqme, bad_point)
    class(prc_gosam_t), intent(in) :: object
    integer, intent(in) :: i_proc
    type(vector4_t), dimension(:), intent(in) :: p
    real(default), intent(in) :: ren_scale
    real(default), intent(in) :: alpha_s
    logical, intent(out) :: bad_point
    real(default), dimension(4), intent(out) :: sqme
    real(double) :: mu
    real(double), dimension(50) :: mom
    real(double), dimension(60) :: r
    character(len=6) :: alpha_par
    integer :: i, k, comp
    real(double) :: acc_dble
    real(default) :: acc
    integer :: ierr
    k = 1
    alpha_par = 'alphaS'
    do i = 1, size (p)
      mom(k:k+3) = vector4_get_components (p(i))
      mom(k+4) = invariant_mass (p(i))
      k = k + 5
      if (k > 50) call msg_fatal ("Too many final-state particles for GoSAM!")
    end do
    mom (k:50) = 0.0
    if (ren_scale == 0.0) then
      mu = sqrt (2* (p(1)*p(2)))
    else
      mu = ren_scale
    end if
    select type (driver => object%driver)
    type is (gosam_driver_t)
      call driver%gosam_olp_set_parameter (alpha_par, &
           dble (twopi), 0._double, ierr)
      call driver%gosam_olp_eval2 (i_proc, mom, mu, r, acc_dble) 
    end select
    acc = acc_dble
    sqme = r(1:4)
    if (acc > object%maximum_accuracy) then
       bad_point = .true.
    else
       bad_point = .false.
    end if
  end subroutine prc_gosam_compute_sqme_fin

  subroutine prc_gosam_fill_constants (object, data_born)
    class(prc_gosam_t), intent(inout) :: object
    type(process_constants_t), intent(in) :: data_born
    associate (data => object%data)
      data%id = 'eeuu'
      print *, 'Model name: ', char (data_born%model_name)
      data%model_name = data_born%model_name
      data%md5sum = data_born%md5sum
      data%openmp_supported = .false.
      data%n_in = data_born%n_in
      data%n_out = data_born%n_out
      data%n_flv = data_born%n_flv
      data%n_hel = data_born%n_hel
      data%n_col = data_born%n_col
      data%n_cin = data_born%n_cin
      data%n_cf = data_born%n_cf
      data%flv_state = data_born%flv_state
      data%hel_state = data_born%hel_state
      data%col_state = data_born%col_state
      data%color_factors = data_born%color_factors
      data%cf_index = data_born%cf_index
      data%ghost_flag = data_born%ghost_flag
    end associate
  end subroutine prc_gosam_fill_constants

  subroutine prc_gosam_allocate_workspace (object, tmp)
    class(prc_gosam_t), intent(in) :: object
    class(workspace_t), intent(inout), allocatable :: tmp
    allocate (gosam_state_t :: tmp)
  end subroutine prc_gosam_allocate_workspace

  function prc_gosam_get_alpha_s (object, tmp) result (alpha)
    class(prc_gosam_t), intent(in) :: object
    class(workspace_t), intent(in), allocatable :: tmp
    real(default) :: alpha
    if (allocated (tmp)) then
      select type (tmp)
      type is (gosam_state_t)
        alpha = tmp%alpha_qcd
      end select
    end if
  end function prc_gosam_get_alpha_s

  subroutine prc_gosam_set_n_proc (object)
    class(prc_gosam_t), intent(inout) :: object
    select type (writer => object%def%writer)
    type is (gosam_writer_t)
      object%n_proc = writer%get_n_proc ()
    end select
  end subroutine prc_gosam_set_n_proc

  function prc_gosam_get_n_proc (object) result (n_proc)
    class(prc_gosam_t), intent(in) :: object
    integer :: n_proc
    n_proc = object%n_proc
  end function prc_gosam_get_n_proc

  subroutine gosam_state_write (object, unit)
    class(gosam_state_t), intent(in) :: object
    integer, intent(in), optional :: unit
    call msg_bug ("gosam_state_write: What to write?")
  end subroutine gosam_state_write

  subroutine gosam_state_reset_new_kinematics (object)
    class(gosam_state_t), intent(inout) :: object
    object%new_kinematics = .true.
  end subroutine gosam_state_reset_new_kinematics

  function check_golem_installation (os_data) result (res)
     type(os_data_t), intent(in) :: os_data
     logical :: res
     type(string_t) :: libavh_olo, libgolem
     integer :: i
     res = .true.
     do i = 1, size (lib_suffix)
       libavh_olo = os_data%prefix // '/lib/libavh_olo' // &
            trim (lib_suffix (i))
       libgolem = os_data%prefix // '/lib/libgolem' // trim (lib_suffix (i))
       res = res .and. (os_file_exist (libavh_olo) .and. &
            os_file_exist (libgolem))
       if (.not. res) exit
     end do
   end function check_golem_installation

  function check_samurai_installation (os_data) result (res)
    type(os_data_t), intent(in) :: os_data
    logical :: res
    type(string_t) :: libsamurai
    integer :: i
    res = .true.
    do i = 1, size (lib_suffix)
      libsamurai = os_data%prefix // '/lib/libsamurai' // trim (lib_suffix(i))
      res = os_file_exist (libsamurai)
      if (.not. res) exit
    end do
  end function check_samurai_installation 

  function check_ninja_installation (os_data) result (res)
    type(os_data_t), intent(in) :: os_data
    logical :: res
    type(string_t) :: libninja
    integer :: i
    res = .true.
    do i = 1, size (lib_suffix)
      libninja = os_data%prefix // '/lib/libninja' // trim (lib_suffix(i))
      res = os_file_exist (libninja)
      if (.not. res) exit
    end do
  end function check_ninja_installation


end module prc_gosam

